;; -*- lisp -*-

(in-package :it.bese.ucw)

;;;; * UCW Request Dispatcher

;;;; Whenever a requset comes into a ucw app we look at its list of
;;;; dispatcher and try to find one which sholud handle the request.

;;;; NB: Dispatcher instances are a shared resource, the dispatch
;;;; function may be called on the same dispatcher object by multiple
;;;; threads. Take this into consideration when writing custom
;;;; dispatcher classes. IOW: don't put any per-request state in the
;;;; dispatcher instance.

(defclass standard-dispatcher ()
  ((priority :accessor priority :initarg :priority
             :initform 0
             :documentation "Dispatchers will be checked from
             highest-priority to lowest priority. The default
             values for priority on the various classes assume
             this is a positive integer less than
             most-positive-fixnum.")))

(defclass function-dispatcher (standard-dispatcher)
  ((handler :accessor handler :initarg :handler
            :documentation "Called when the dispatchers finds a
            maching request. This function will be passed the app
            and the context."))
  (:documentation "Superclass of all dispatchers whcih call some
  user supplied function."))

(defgeneric dispatch (dispatcher application context)
  (:documentation "Entry point into a dispatcher. Must return T
  if the context has been handled or NIL if it hasn't.

No methods defined on this function may rebind *context*, nor
change CONTEXT's application. Only if the method returns T is it
allowed to modify CONTEXT or APPLICATION, even in that case
methods defined on this function must not modify CONTEXT's
application nor rebind *context*."))

;;;; ** URL based dispatchers

;;;; URL-DISPATCHER and REGEXP-DISPATCHER call a block of code based
;;;; on the query-path. The macro URL-DISPATCHER and REGEXP-DISPATCHER
;;;; provide make-instance form make it easy to tie some action code
;;;; to a particular url.

(defclass url-dispatcher (function-dispatcher)
  ((url-string :accessor url-string :initarg :url-string))
  (:documentation "Calls some code when the url of the request
  matches URL-STRING.

By default this class sets up the session and frame machansims
neccessary for backtracking and continuations.")
  (:default-initargs :priority 1))

(defmacro url-dispatcher (url-string &body body)
  "Returns a dispatcher which matches when URL-STRING matches and
executes, as if in a defaction, BODY."
  `(make-instance 'url-dispatcher
                  :url-string ,url-string
                  :handler (lambda (application context)
                             (with-call/cc
                               (let ((self nil))
                                 ,@body)))))

(defmethod print-object ((dispatcher url-dispatcher) stream)
  (print-unreadable-object (dispatcher stream :type t :identity t)
    (format stream "~S" (url-string dispatcher))))

(defmethod dispatch :before ((dispatcher url-dispatcher)
                             (application standard-application)
                             (context standard-request-context))
  (ucw.rerl.dispatcher.dribble "Attempting to match ~S against ~S."
                               (query-path-sans-prefix context)
                               (url-string dispatcher)))

(defmethod dispatch ((dispatcher url-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  (when (string= (url-string dispatcher) (query-path-sans-prefix context))
    (ucw.rerl.dispatcher.info "URL-DISPATCHER matched, passing control to ~S." (handler dispatcher))
    (ensure-session application context)
    (handle-action-as-ucw-action (lambda ()
                                   (funcall (handler dispatcher) application context)))
    (flush-request-response)
    t))

(defclass regexp-dispatcher (url-dispatcher)
  ())

(defvar *dispatcher-registers* nil
  "The second value retruned by cl-ppcre:scan-to-strings is bound
  to this variable during the execution of the action.")

(defmethod dispatch ((dispatcher regexp-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  (multiple-value-bind (matchp *dispatcher-registers*)
      (cl-ppcre:scan-to-strings (url-string dispatcher)
                                (query-path-sans-prefix context)
                                :sharedp t)
    (when matchp
      (ucw.rerl.dispatcher.info "URL-DISPATCHER matched, passing control to ~S." (handler dispatcher))
      (ensure-session application context)
      (handle-action-as-ucw-action
       (lambda ()
         (funcall (handler dispatcher) application context)))
      (flush-request-response)
      t)))

(defmacro regexp-dispatcher (url-regexp &body body)
  "Returns a dispatcher which matches when URL-REGEXP (a regular
expression which is passed untouched to cl-ppcre:scan-to-strings)
matches and executes, as if in a defaction, BODY.

Any registers in URL-REGEXP are available through the array bound
to *dispatcher-registers*."
  `(make-instance 'regexp-dispatcher
                  :url-string ,url-regexp
                  :handler (lambda (application context)
                             (with-call/cc
                               (let ((self nil))
                                 ,@body)))))

;;;; *** Tal Directory dispatcher

(defclass tal-dispatcher (standard-dispatcher)
  ()
  (:documentation "This dispatcher simply serves up TAL pages.

Tal pages can refer to the session object (and it will be
maintained across requests) but must avoid using call/answer or
actions.")
  (:default-initargs :priority 0))

(defmethod dispatch ((dispatcher tal-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  ;; see if the required tal template exists
  (when-bind template-truename (template-truename (application.tal-generator application)
                                                  (make-pathname :type "tal"
                                                                 :defaults (query-path-sans-prefix context)))
    (ucw.rerl.dispatcher.info "Publishing tal ~S as ~S." template-truename (query-path-sans-prefix context))
    (ensure-session application context)
    (make-new-frame (context.session context))
    (block render
      (handler-bind ((error (lambda (condition)
                              (catch 'abort-action
                                (handle-action-error condition (collect-backtrace condition)))
                              (render-loop (frame.window-component (context.current-frame *context*)))
                              (return-from render nil))))
        (%render-template context
                          (application.tal-generator application)
                          template-truename
                          'nil)))
    (flush-request-response)
    t))

;;;; ** ACTION-DISPATCHER

;;;; This is the core dispatcher for ucw. Due to how ucw's compute-url
;;;; method works it is important that the action-dispatcher be
;;;; checked before any url-dispatchers, so it sholud be the first
;;;; element in an application's dispatcher list.

(defclass action-dispatcher (standard-dispatcher)
  ()
  (:documentation "If the context contains enough info to call an
  action we call it, otherwise we bail.")
  (:default-initargs :priority most-positive-fixnum))

(defmethod dispatch ((dispatcher action-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  (when-bind session (find-session application context)
    (ensure-session application context session)
    (ucw.rerl.dispatcher.dribble "Found session ~S." session)
    (when-bind frame-id (find-frame-id context)
      (ucw.rerl.dispatcher.dribble "Found frame-id ~S." frame-id)
      (when-bind frame (find-frame-by-id session frame-id)
        (ucw.rerl.dispatcher.dribble "Found frame ~S." frame)
        (when-bind action (find-action frame (find-action-id context))
          (ucw.rerl.dispatcher.dribble "ACTION-DISPATCHER matched, calling action ~S." action)
          (setf (session.current-frame session) frame)
          (reinstate-backtracked frame)
          (call-callbacks frame (context.request context))
          (handle-action-as-ucw-action action)
          (flush-request-response)
          t)))))

;;;; ** Simple, REST-ish, dispatchers

(defclass minimal-dispatcher (regexp-dispatcher)
  ()
  (:documentation "A dispatcher which does as little work as
  possible. The handler function must do basically
  everything (including shutdowning down reuest and response)."))

(defmethod dispatch ((dispatcher minimal-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  (multiple-value-bind (matchp *dispatcher-registers*)
      (cl-ppcre:scan-to-strings (url-string dispatcher)
                                (query-path-sans-prefix context)
                                :sharedp t)
    (when matchp
      (ucw.rerl.dispatcher.info "MINIMAL-DISPATCHER matched, passing control to ~S." (handler dispatcher))
      (funcall (handler dispatcher) application context)
      t)))

(defclass simple-dispatcher (minimal-dispatcher)
  ()
  (:documentation "This class of dispatchers avoids all of UCW's
  standard call/cc (and therefore frame/backtracking/component)
  mechanism.

Unlike all other UCW dispatchers a simple-dispatcher must not use
CALL, and must perform the rendering directly within the handler."))

(defmethod dispatch ((dispatcher simple-dispatcher)
                     (application standard-application)
                     (context standard-request-context))
  (multiple-value-bind (matchp *dispatcher-registers*)
      (cl-ppcre:scan-to-strings (url-string dispatcher)
                                (query-path-sans-prefix context)
                                :sharedp t)
    (when matchp
      (ucw.rerl.dispatcher.info "SIMPLE-DISPATCHER matched, passing control to ~S." (handler dispatcher))
      (ensure-session application context)
      (funcall (handler dispatcher) application context)
      (flush-request-response)
      t)))

(defmacro simple-dispatcher (url-regexp &body body)
  `(make-instance 'simple-dispatcher
                  :url-string ,url-regexp
                  :handler (lambda (application context)
                             (declare (ignorable application context))
                             ,@body)))

(defclass parenscript-dispatcher (minimal-dispatcher)
  ((cached-javascript :accessor cached-javascript)
   (last-compiled :accessor last-compiled)
   (parenscript-file :accessor parenscript-file :initarg :parenscript-file))
  (:documentation "This derived regex dispatcher serves a compiled
javascript string from the given PARENSCRIPT-FILE.

The file is compiled lazily on the first request and in-package
forms are working as expected. Each form is evaluated, so
using backquote is possible and in fact a must. The following
dynamic variables are bound then compiling:
  *context* - the current request context (hint: (context.application *context*))"))

(defun parenscript-dispatcher (url-regex parenscript-file &key (cache t))
  "Return a parenscript-dispatcher"
  (let ((dispatcher (make-instance 'parenscript-dispatcher
                                   :url-string url-regex
                                   :parenscript-file parenscript-file)))
    (setf (handler dispatcher)
          (lambda (application context)
            (declare (ignore application context))
            (when (or (not cache)
                      (not (slot-boundp dispatcher 'cached-javascript))
                      (> (file-write-date (parenscript-file dispatcher))
                         (last-compiled dispatcher)))
              (setf (cached-javascript dispatcher)
                    (js:compile-parenscript-file-to-string (parenscript-file dispatcher)))
              (setf (last-compiled dispatcher) (get-universal-time)))
            ;; TODO make sure some cache info is sent so that it will be cached by the browser
            (serve-sequence (cached-javascript dispatcher)
                            :last-modified (last-compiled dispatcher)
                            :content-type "text/javascript")))
    dispatcher))


;;;; ** Utility Functions for dispatchers

;;;; These functions wrap up some of the common things dispatchers
;;;; must do. They could probably be replaced with a more complex
;;;; class hierarchy and methods on dispatch, but I (mb) do not
;;;; believe that would actually make things clearer.

(defun flush-request-response (&optional (context *context*))
  "Shutdown the reuest and send the response's data.

FLUSH-REQUEST-RESPONSE always returns T, this allows it to be
used as the last form in a dispatcher."
  (close-request (context.request context))
  (send-response (context.response context))
  t)

(defun ensure-session (application context &optional (session (or (find-session application context)
                                                                  (make-new-session application))))
  "If CONTEXT's request specifies a session then put it in the
  context, otherwise create a new context.

Updates the session's last-access time."
  (setf (context.session context) session
        (session.last-access (context.session context)) (get-universal-time)))

(defun handle-action-as-ucw-action (action)
  (make-new-frame (context.session *context*))
  (with-action-error-handler ()
    (funcall action))
  ;; nb: don't use the same frame as returned by make-new-frame
  ;; to get the current component. 99% of the time they will be
  ;; the same as below, but 1% they will be differenet and that
  ;; 1% of the time is when all hell has broken lose and we're
  ;; dancing with cats...
  (render-loop (frame.window-component (context.current-frame *context*)))
  (save-backtracked (context.current-frame *context*)))

;; Copyright (c) 2003-2006 Edward Marco Baringer
;; All rights reserved. 
;; 
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;; 
;;  - Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 
;;  - Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 
;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;    of its contributors may be used to endorse or promote products
;;    derived from this software without specific prior written permission.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
